//   Copyright (c)  2011  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/decimal.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/QQ.H"
#include "CoCoA/ZZ.H"

#include <algorithm>
using std::fill;
using std::copy;
#include <iostream>
using std::ostream;
//#include <string>
using std::string;
#include <sstream>
using std::ostringstream;

namespace CoCoA
{

  std::ostream& operator<<(std::ostream& out, const MantExp& ME)
  {
    out << "MantExp(s=" << ME.sign << ", e=" << ME.exponent << ", m=" << ME.mantissa << ")";
    return out;
  }


  MantExp MantissaAndExponent(const ZZ& N, const MachineInteger& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "MantissaAndExponent");
    if (IsZero(N)) return MantExp(0,0,ZZ(0));
    const long ndigits = AsSignedLong(SigFig);
    const int SignN = sign(N);
    const long e = ILogBase(N,10); // overflow???
    if (e < ndigits)
      return MantExp(SignN, e, abs(N)*power(10, ndigits-e-1));
    const ZZ HalfULP = 5*power(10, e-ndigits);
    const ZZ digits = (1+abs(N)/HalfULP)/2;
    // Must check whether digits has overflowed...
    if (abs(digits) == power(10, ndigits))
      return MantExp(SignN, e+1, digits/10);
    return MantExp(SignN, e, digits);
  }


  MantExp MantissaAndExponent(const QQ& q, const MachineInteger& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "MantissaAndExponent");
    if (IsZero(q)) return MantExp(0,0,ZZ(0));
    if (IsOneDen(q)) return MantissaAndExponent(num(q), SigFig);
    const long ndigits = AsSignedLong(SigFig);
    const int signq = sign(q);
    const long e = ILogBase(q,10); // overflow???
    ZZ digits;
    if (e < ndigits)
      digits = round(abs(q)*power(10,ndigits-e-1));
    else
      digits = round(abs(q)/power(10,1+e-ndigits)); 
    // Must check whether digits has overflowed...
    if (abs(digits) == power(10, ndigits))
      return MantExp(signq, e+1, digits/10);
    return MantExp(signq, e, digits);
  }


  std::string FloatStr(const MantExp& ME, long ndigits)
  {
    string mantissa; mantissa.reserve(ndigits);
    if (!IsZero(ME.mantissa))
      IsConvertible(mantissa, ME.mantissa); // cannot fail, except std::bad_alloc
    else
      // fill string with ndigits of zeroes...
      for (long i=0;i<ndigits;++i)mantissa+='0';///???fill(...);

    string ans; ans.reserve(9+ndigits+10); // 9 = size of "-.*10^(-)", 10 is for 32-bit exponent
// MIGHT BE NEATER USING A     back_inserter(ans); ???
    if (ME.sign < 0) ans += '-';
    ans += mantissa[0];
    ans += '.';
    copy(&mantissa[1], &mantissa[ndigits], back_inserter(ans));
    ans += "*10^";
    if (ME.exponent < 0) ans += '(';
    ostringstream exp;
    exp << ME.exponent;
    ans += exp.str();
    if (ME.exponent < 0) ans += ')';
    return ans;
  }

  std::string FloatStr(const ZZ& N, const MachineInteger& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "FloatStr");
    return FloatStr(MantissaAndExponent(N, SigFig), AsSignedLong(SigFig));
  }

  std::string FloatStr(const QQ& q, const MachineInteger& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "FloatStr");
    return FloatStr(MantissaAndExponent(q, SigFig), AsSignedLong(SigFig));
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/decimal.C,v 1.1 2011/03/01 15:21:22 abbott Exp $
// $Log: decimal.C,v $
// Revision 1.1  2011/03/01 15:21:22  abbott
// Added new functions: MantissaAndExponent, FloatStr (in CoCoALib).
//
//
